/* MCU specific files */
#include "common.h"
#include "mc.h"
#include "pll.h"
#include "uart.h"

/* FreeRTOS includes */
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include "semphr.h"

/* GHS specific symbols which provides addresses of stack and data sections */
extern char __ghsbegin_sdabase[];
extern char __ghsend_stack[];
extern char __smt_s_vectStart[];
extern char __smt_h_vectStart[];

void _start( void );

//make sure the micro will execute proper code after reset
#if (EXECUTEFROM == LOCATION_FLASH)

    //Set RCHW word to indicate valid boot location
    #pragma ghs section rodata=".rchw"
    const UINT32 rchwValue = 0x005A005A;    //to enable watch dog set bit 5 (WTE) to one. i.e. |0x04000000
                                            //to indicate the code is VLE code set bit 7 (VLE) to one. i.e. |0x01000000
    //pointer to the first instruction of the application code
    #pragma ghs section rodata=".jmpaddress"
    const UINT32 startAddress = (UINT32) _start;

    #pragma ghs section rodata=default
#endif

    //Startup code - this is the first code which is executed after reset
    //               it initalizes stack, RAM, sets MSR register and calls main()
    #pragma ghs section text=".start"
    #pragma asm
        .vle
        .section ".start", "vax"
        .globl _start

        _start:
            ;initialize stack
            e_lis   r1, __ghsend_stack@h
            e_or2i  r1, __ghsend_stack@l

            ;initialize RAM memory to avoid CRC exceptions
            ;(Only used/occupied memory area is initialized)
    #ifdef INIT_RAM
            e_lis   r11, __ghsbegin_sdabase@h
            e_or2i  r11, __ghsbegin_sdabase@l
            e_lis   r12, __ghsend_stack@h
            e_or2i  r12, __ghsend_stack@l

        init_ram_loop:
            e_stmw  r0,  0(r11)
            e_addi  r11, r11, 128
            cmpw    cr0, r11, r12
            e_blt   cr0, init_ram_loop
    #endif

            ;Fill in .data and .sdata sections from their flash copy
            ;(This is valid only when locate_standalone_flash.ld and
            ; initialized variables are going to be used)
            ; It's programmer's responsibility to decide whether this is needed or not
    #ifdef INIT_RAM_VARIABLES
            ;prepare source address
            e_lis   r11, __smt_initVariablesROMCopyStart@h
            e_or2i  r11, __smt_initVariablesROMCopyStart@l

            ;prepare source end address
            e_lis   r12, __smt_initVariablesROMCopyEnd@h
            e_or2i  r12, __smt_initVariablesROMCopyEnd@l

            ;prepare destination address
            e_lis   r10, __smt_initVariablesStart@h
            e_or2i  r10, __smt_initVariablesStart@l

        ram_copy_loop:
            cmpw    cr0, r11, r12
            e_beq   cr0, ram_copy_end

            e_lbzu  r4,  0(r11)
            e_stb   r4,  0(r10)
            e_addi  r11, r11, 1
            e_addi  r10, r10, 1
            e_b     ram_copy_loop
        ram_copy_end:
    #endif
            ;jump to main function
            e_bl   main
    #pragma endasm
    #pragma ghs section text=default
/*---------------------------------------------------------------------------*/
#define IVPR                                63 // IVPR SPR number
#define timerPERIOD                     640000 // 1s with prescaler 1:100

/* Tasks priorities */
#define mainFIRST_TASK_PRIORITY              ( tskIDLE_PRIORITY + 1 )
#define mainSECOND_TASK_PRIORITY             ( tskIDLE_PRIORITY + 3 )

#define DEBUG(x) \
    UART_PutChar('['); UART_PutInteger(xTaskGetTickCount()); UART_PutString("]:"); \
    UART_PutString(x);

volatile portLONG secCounter = 0;  // helpers variable for interrupt handler
volatile portLONG msgCounter = 0;  // counter of received messages
portCHAR arrMessages[] = {'A', 'B', 'C', 'D'};
int msgPointer = 0;
xQueueHandle xMsgQueue = 0;

/*---------------------------------------------------------------------------*/
/* STM0 interrupt handler */
void STM0_InterruptHandler( void )
{
    portBASE_TYPE xHigherPriorityTaskWoken = pdTRUE;

    STM.CMP0.R = STM.CNT0.R + timerPERIOD;  // new compare
    STM.CIR0.R = 1;                         // clear flag

    if ( !(secCounter++ % 10) ) {
        xQueueSendFromISR( xMsgQueue,
            &arrMessages[msgPointer++ % sizeof (arrMessages)],
            &xHigherPriorityTaskWoken );
    }
}
/*---------------------------------------------------------------------------*/

/* Demo tasks */
portTASK_FUNCTION( vTaskOne, pvParameters )
{
    for (;;) {
        DEBUG( "(task #1) -- Test message --\n" );
        vTaskDelay( 1000 );
    }
}
/*---------------------------------------------------------------------------*/

portTASK_FUNCTION( vTaskTwo, pvParameters )
{
    for ( ;; ) {
        char message= 0;
        if ( xQueueReceive( xMsgQueue, &message, 50 ) == pdTRUE ) {


            DEBUG( "(task #2) Received[" );
            UART_PutInteger( msgCounter++ );
            UART_PutString( "] -> " );
            UART_PutChar( message );
            UART_PutChar( '\n' );

            vTaskDelay( 100 );
        }
    }
}

/*
 * Performs hardware modules initialization
 */
static void prvHardwareSetup( void );
/*---------------------------------------------------------------------------*/

void main( void )
{
    /* Hardware initialization code */
    prvHardwareSetup();

    /* Creating queue for messages from ST0 interrupt handler */
    DEBUG( "* Creating message queue..." );
    xMsgQueue= xQueueCreate( 5, sizeof (portCHAR) );
    if ( xMsgQueue )
        UART_PutString( "done\n" );
    else
        UART_PutString( "fail\n" );

    /* Creating first task for messages polling */
    DEBUG( "* Creating task #1..." );
    if ( xTaskCreate( vTaskOne, "task #1", configMINIMAL_STACK_SIZE,
             NULL, mainFIRST_TASK_PRIORITY, NULL ) == pdPASS )
        UART_PutString( "done\n" );
    else
        UART_PutString( "fail\n" );

    /* Creating second task */
    DEBUG( "* Creating task #2..." );
    if ( xTaskCreate( vTaskTwo, "task #2", configMINIMAL_STACK_SIZE,
             NULL, mainSECOND_TASK_PRIORITY, NULL ) == pdPASS )
        UART_PutString( "done\n" );
    else
        UART_PutString( "fail\n" );

    DEBUG( "* FreeRTOS started!\n" );

    /* Start the FreeRTOS scheduler */
    vTaskStartScheduler();

    /* Should never reach here */
    for ( ;; )
        ;
}
/*---------------------------------------------------------------------------*/

void prvHardwareSetup( void )
{
    /* Configure interrupt controller */
    INTC.MCR.R   = 0;   // set software vector mode, default table entry offset of 4 bytes
    __MTSPR( IVPR, (uint32_t)__smt_s_vectStart); // setting address of core vectors
    INTC.IACKR.R = (uint32_t)__smt_h_vectStart;  // setting address of hardware vectors
    INTC.CPR.R   = 0;   // enable priorities

    /* Disable SoftwareWatchdogTimer */
    SWT.SR.R        = 0xC520;
    SWT.SR.R        = 0xD928; // unlock peripheral
    SWT.CR.B.WEN    = 0;

    /* Configure RUN0 mode, with:
     a) FXOSC ON
     b) PLL ON
     c) SYSCLK=PLL
    */
    ME.RUN[0].R |= 0x74;

    /* Configure Run Peripheral Configuration register in order to active peripherals */
    ME.RUNPC[0].R   = 0x000000FE;

    /* Setup system PLL (PLL0) */
    CGM.FMPLL[0].CR.R = (FMPLL_IDF << 26 | FMPLL_ODF << 24 | FMPLL_NDIV << 16) & 0x3F7F0000;

    /* Enter RUN0 Mode */
    me_modeTransition(ME_MODE_RUN0);

    /* Setting up STM0 timer */
    STM.CR0.R = 3;             // setup STM with freeze
    STM.CR0.B.CPS = 99;        // div 100
    STM.CMP0.R = STM.CNT0.R + timerPERIOD;    // first start
    STM.CCR0.R = 1;            // enable channel 0
    INTC.PSR[30].R = 1;        // set priority

    UART_Init();
}
/*---------------------------------------------------------------------------*/
